@@ -5,12 +5,12 @@
  *	  commands.  At one time acted as an interface between the Lisp and C
  *	  systems.
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/tcop/utility.c,v 1.125.2.3 2005/01/24 17:46:58 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/tcop/utility.c,v 1.180.2.1 2005/01/24 17:46:49 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -18,34 +18,40 @@
 
 #include "access/heapam.h"
 #include "catalog/catalog.h"
+#include "catalog/namespace.h"
 #include "catalog/pg_shadow.h"
 #include "commands/async.h"
 #include "commands/cluster.h"
-#include "commands/command.h"
 #include "commands/comment.h"
 #include "commands/copy.h"
-#include "commands/creatinh.h"
+#include "commands/conversioncmds.h"
 #include "commands/dbcommands.h"
 #include "commands/defrem.h"
 #include "commands/explain.h"
+#include "commands/lockcmds.h"
+#include "commands/portalcmds.h"
+#include "commands/prepare.h"
 #include "commands/proclang.h"
-#include "commands/rename.h"
+#include "commands/schemacmds.h"
 #include "commands/sequence.h"
+#include "commands/tablecmds.h"
 #include "commands/trigger.h"
 #include "commands/user.h"
 #include "commands/vacuum.h"
-#include "commands/variable.h"
 #include "commands/view.h"
 #include "miscadmin.h"
+#include "nodes/makefuncs.h"
 #include "parser/parse.h"
 #include "parser/parse_clause.h"
 #include "parser/parse_expr.h"
+#include "parser/parse_type.h"
 #include "rewrite/rewriteDefine.h"
 #include "rewrite/rewriteRemove.h"
 #include "tcop/utility.h"
 #include "utils/acl.h"
+#include "utils/guc.h"
+#include "utils/lsyscache.h"
 #include "utils/syscache.h"
-#include "utils/temprel.h"
 #include "access/xlog.h"
 
 /*
@@ -65,6 +71,7 @@ static struct kindstrings kindstringarray[] = {
 	{RELKIND_SEQUENCE, "a", "sequence", "SEQUENCE"},
 	{RELKIND_VIEW, "a", "view", "VIEW"},
 	{RELKIND_INDEX, "an", "index", "INDEX"},
+	{RELKIND_COMPOSITE_TYPE, "a", "type", "TYPE"},
 	{'\0', "a", "???", "???"}
 };
 
@@ -94,9 +101,10 @@ DropErrorMsg(char *relname, char wrongkind, char rightkind)
 }
 
 static void
-CheckDropPermissions(char *name, char rightkind)
+CheckDropPermissions(RangeVar *rel, char rightkind)
 {
 	struct kindstrings *rentry;
+	Oid			relOid;
 	HeapTuple	tuple;
 	Form_pg_class classform;
 
@@ -105,25 +113,55 @@ CheckDropPermissions(char *name, char rightkind)
 			break;
 	Assert(rentry->kind != '\0');
 
-	tuple = SearchSysCache(RELNAME,
-						   PointerGetDatum(name),
+	relOid = RangeVarGetRelid(rel, true);
+	if (!OidIsValid(relOid))
+		elog(ERROR, "%s \"%s\" does not exist", rentry->name, rel->relname);
+	tuple = SearchSysCache(RELOID,
+						   ObjectIdGetDatum(relOid),
 						   0, 0, 0);
 	if (!HeapTupleIsValid(tuple))
-		elog(ERROR, "%s \"%s\" does not exist", rentry->name, name);
+		elog(ERROR, "%s \"%s\" does not exist", rentry->name, rel->relname);
 
 	classform = (Form_pg_class) GETSTRUCT(tuple);
 
 	if (classform->relkind != rightkind)
-		DropErrorMsg(name, classform->relkind, rightkind);
+		DropErrorMsg(rel->relname, classform->relkind, rightkind);
 
-	if (!pg_ownercheck(GetUserId(), name, RELNAME))
-		elog(ERROR, "you do not own %s \"%s\"",
-			 rentry->name, name);
+	/* Allow DROP to either table owner or schema owner */
+	if (!pg_class_ownercheck(relOid, GetUserId()) &&
+		!pg_namespace_ownercheck(classform->relnamespace, GetUserId()))
+		aclcheck_error(ACLCHECK_NOT_OWNER, rel->relname);
 
-	if (!allowSystemTableMods && IsSystemRelationName(name) &&
-		!is_temp_relname(name))
+	if (!allowSystemTableMods && IsSystemClass(classform))
 		elog(ERROR, "%s \"%s\" is a system %s",
-			 rentry->name, name, rentry->name);
+			 rentry->name, rel->relname, rentry->name);
+
+	ReleaseSysCache(tuple);
+}
+
+static void
+CheckOwnership(RangeVar *rel, bool noCatalogs)
+{
+	Oid			relOid;
+	HeapTuple	tuple;
+
+	relOid = RangeVarGetRelid(rel, false);
+	tuple = SearchSysCache(RELOID,
+						   ObjectIdGetDatum(relOid),
+						   0, 0, 0);
+	if (!HeapTupleIsValid(tuple))
+		elog(ERROR, "Relation \"%s\" does not exist", rel->relname);
+
+	if (!pg_class_ownercheck(relOid, GetUserId()))
+		aclcheck_error(ACLCHECK_NOT_OWNER, rel->relname);
+
+	if (noCatalogs)
+	{
+		if (!allowSystemTableMods &&
+			IsSystemClass((Form_pg_class) GETSTRUCT(tuple)))
+			elog(ERROR, "relation \"%s\" is a system catalog",
+				 rel->relname);
+	}
 
 	ReleaseSysCache(tuple);
 }
@@ -149,7 +187,6 @@ ProcessUtility(Node *parsetree,
 			   char *completionTag)
 {
 	char	   *relname;
-	char	   *relationName;
 
 	if (completionTag)
 		completionTag[0] = '\0';
@@ -170,6 +207,29 @@ ProcessUtility(Node *parsetree,
 						BeginTransactionBlock();
 						break;
 
+						/*
+						 * START TRANSACTION, as defined by SQL99:
+						 * Identical to BEGIN, except that it takes a few
+						 * additional options.
+						 */
+					case START:
+						{
+							BeginTransactionBlock();
+
+							/*
+							 * Currently, the only option that can be set
+							 * by START TRANSACTION is the isolation
+							 * level.
+							 */
+							if (stmt->options)
+							{
+								SetPGVariable("TRANSACTION ISOLATION LEVEL",
+											  stmt->options,
+											  false);
+							}
+						}
+						break;
+
 					case COMMIT:
 						EndTransactionBlock();
 						break;
@@ -182,8 +242,7 @@ ProcessUtility(Node *parsetree,
 			break;
 
 			/*
-			 * ******************************** portal manipulation ********************************
-			 *
+			 * ************************* portal manipulation ***************************
 			 */
 		case T_ClosePortalStmt:
 			{
@@ -200,8 +259,6 @@ ProcessUtility(Node *parsetree,
 				bool		forward;
 				int			count;
 
-				SetQuerySnapshot();
-
 				forward = (bool) (stmt->direction == FORWARD);
 
 				/*
@@ -216,142 +273,126 @@ ProcessUtility(Node *parsetree,
 			break;
 
 			/*
-			 * ******************************** relation and attribute
-			 * manipulation ********************************
-			 *
+			 * relation and attribute manipulation
 			 */
-		case T_CreateStmt:
-			DefineRelation((CreateStmt *) parsetree, RELKIND_RELATION);
+		case T_CreateSchemaStmt:
+			CreateSchemaCommand((CreateSchemaStmt *) parsetree);
+			break;
 
-			/*
-			 * Let AlterTableCreateToastTable decide if this one needs a
-			 * secondary relation too.
-			 */
-			CommandCounterIncrement();
-			AlterTableCreateToastTable(((CreateStmt *) parsetree)->relname,
-									   true);
+		case T_CreateStmt:
+			{
+				Oid			relOid;
+
+				relOid = DefineRelation((CreateStmt *) parsetree,
+										RELKIND_RELATION);
+
+				/*
+				 * Let AlterTableCreateToastTable decide if this one needs
+				 * a secondary relation too.
+				 */
+				CommandCounterIncrement();
+				AlterTableCreateToastTable(relOid, true);
+			}
 			break;
 
 		case T_DropStmt:
 			{
 				DropStmt   *stmt = (DropStmt *) parsetree;
-				List	   *args = stmt->names;
 				List	   *arg;
 
-				foreach(arg, args)
+				foreach(arg, stmt->objects)
 				{
-					relname = strVal(lfirst(arg));
+					List	   *names = (List *) lfirst(arg);
+					RangeVar   *rel;
 
 					switch (stmt->removeType)
 					{
 						case DROP_TABLE:
-							CheckDropPermissions(relname, RELKIND_RELATION);
-							RemoveRelation(relname);
+							rel = makeRangeVarFromNameList(names);
+							CheckDropPermissions(rel, RELKIND_RELATION);
+							RemoveRelation(rel, stmt->behavior);
 							break;
 
 						case DROP_SEQUENCE:
-							CheckDropPermissions(relname, RELKIND_SEQUENCE);
-							RemoveRelation(relname);
+							rel = makeRangeVarFromNameList(names);
+							CheckDropPermissions(rel, RELKIND_SEQUENCE);
+							RemoveRelation(rel, stmt->behavior);
 							break;
 
 						case DROP_VIEW:
-							CheckDropPermissions(relname, RELKIND_VIEW);
-							RemoveView(relname);
+							rel = makeRangeVarFromNameList(names);
+							CheckDropPermissions(rel, RELKIND_VIEW);
+							RemoveView(rel, stmt->behavior);
 							break;
 
 						case DROP_INDEX:
-							CheckDropPermissions(relname, RELKIND_INDEX);
-							RemoveIndex(relname);
+							rel = makeRangeVarFromNameList(names);
+							CheckDropPermissions(rel, RELKIND_INDEX);
+							RemoveIndex(rel, stmt->behavior);
 							break;
 
-						case DROP_RULE:
-							{
-								char	   *rulename = relname;
-								int			aclcheck_result;
-
-								relationName = RewriteGetRuleEventRel(rulename);
-								aclcheck_result = pg_aclcheck(relationName, GetUserId(), ACL_RULE);
-								if (aclcheck_result != ACLCHECK_OK)
-									elog(ERROR, "%s: %s", relationName,
-										 aclcheck_error_strings[aclcheck_result]);
-								RemoveRewriteRule(rulename);
-							}
-							break;
-
-						case DROP_TYPE_P:
+						case DROP_TYPE:
 							/* RemoveType does its own permissions checks */
-							RemoveType(relname);
+							RemoveType(names, stmt->behavior);
+							break;
+
+						case DROP_DOMAIN:
+
+							/*
+							 * RemoveDomain does its own permissions
+							 * checks
+							 */
+							RemoveDomain(names, stmt->behavior);
+							break;
+
+						case DROP_CONVERSION:
+							DropConversionCommand(names, stmt->behavior);
+							break;
+
+						case DROP_SCHEMA:
+
+							/*
+							 * RemoveSchema does its own permissions
+							 * checks
+							 */
+							RemoveSchema(names, stmt->behavior);
 							break;
 					}
 
 					/*
-					 * Make sure subsequent loop iterations will see
-					 * results of this one; needed if removing multiple
-					 * rules for same table, for example.
+					 * We used to need to do CommandCounterIncrement()
+					 * here, but now it's done inside performDeletion().
 					 */
-					CommandCounterIncrement();
 				}
 			}
 			break;
 
 		case T_TruncateStmt:
 			{
-				Relation	rel;
+				TruncateStmt *stmt = (TruncateStmt *) parsetree;
 
-				relname = ((TruncateStmt *) parsetree)->relName;
-				if (!allowSystemTableMods && IsSystemRelationName(relname))
-					elog(ERROR, "TRUNCATE cannot be used on system tables. '%s' is a system table",
-						 relname);
-
-				/* Grab exclusive lock in preparation for truncate... */
-				rel = heap_openr(relname, AccessExclusiveLock);
-				if (rel->rd_rel->relkind == RELKIND_SEQUENCE)
-					elog(ERROR, "TRUNCATE cannot be used on sequences. '%s' is a sequence",
-						 relname);
-				if (rel->rd_rel->relkind == RELKIND_VIEW)
-					elog(ERROR, "TRUNCATE cannot be used on views. '%s' is a view",
-						 relname);
-				heap_close(rel, NoLock);
-
-				if (!pg_ownercheck(GetUserId(), relname, RELNAME))
-					elog(ERROR, "you do not own class \"%s\"", relname);
-				TruncateRelation(relname);
+				TruncateRelation(stmt->relation);
 			}
 			break;
 
 		case T_CommentStmt:
-			{
-				CommentStmt *statement;
-
-				statement = ((CommentStmt *) parsetree);
-
-				CommentObject(statement->objtype, statement->objname,
-							  statement->objproperty, statement->objlist,
-							  statement->comment);
-			}
+			CommentObject((CommentStmt *) parsetree);
 			break;
 
 		case T_CopyStmt:
-			{
-				CopyStmt   *stmt = (CopyStmt *) parsetree;
+			DoCopy((CopyStmt *) parsetree);
+			break;
 
-				if (stmt->direction != FROM)
-					SetQuerySnapshot();
+		case T_PrepareStmt:
+			PrepareQuery((PrepareStmt *) parsetree);
+			break;
 
-				DoCopy(stmt->relname,
-					   stmt->binary,
-					   stmt->oids,
-					   (bool) (stmt->direction == FROM),
-					   (bool) (stmt->filename == NULL),
+		case T_ExecuteStmt:
+			ExecuteQuery((ExecuteStmt *) parsetree, dest);
+			break;
 
-				/*
-				 * null filename means copy to/from stdout/stdin, rather
-				 * than to/from a file.
-				 */
-					   stmt->filename,
-					   stmt->delimiter,
-					   stmt->null_print);
-			}
+		case T_DeallocateStmt:
+			DeallocateQuery((DeallocateStmt *) parsetree);
 			break;
 
 			/*
@@ -360,43 +401,53 @@ ProcessUtility(Node *parsetree,
 		case T_RenameStmt:
 			{
 				RenameStmt *stmt = (RenameStmt *) parsetree;
+				Oid			relid;
 
-				relname = stmt->relname;
-				if (!allowSystemTableMods && IsSystemRelationName(relname))
-					elog(ERROR, "ALTER TABLE: relation \"%s\" is a system catalog",
-						 relname);
-				if (!pg_ownercheck(GetUserId(), relname, RELNAME))
-					elog(ERROR, "permission denied");
+				CheckOwnership(stmt->relation, true);
 
-				/* ----------------
-				 *	XXX using len == 3 to tell the difference
-				 *		between "rename rel to newrel" and
-				 *		"rename att in rel to newatt" will not
-				 *		work soon because "rename type/operator/rule"
-				 *		stuff is being added. - cim 10/24/90
-				 * ----------------
-				 * [another piece of amuzing but useless anecdote -- ay]
-				 */
-				if (stmt->column == NULL)
+				relid = RangeVarGetRelid(stmt->relation, false);
+
+				switch (stmt->renameType)
 				{
-					/*
-					 * rename relation
-					 *
-					 * Note: we also rename the "type" tuple corresponding to
-					 * the relation.
-					 */
-					renamerel(relname,	/* old name */
-							  stmt->newname);	/* new name */
-				}
-				else
-				{
-					/*
-					 * rename attribute
-					 */
-					renameatt(relname,	/* relname */
-							  stmt->column,		/* old att name */
-							  stmt->newname,	/* new att name */
-							  interpretInhOption(stmt->inhOpt));		/* recursive? */
+					case RENAME_TABLE:
+						{
+							/*
+							 * RENAME TABLE requires that we (still) hold
+							 * CREATE rights on the containing namespace,
+							 * as well as ownership of the table.
+							 */
+							Oid			namespaceId = get_rel_namespace(relid);
+							AclResult	aclresult;
+
+							aclresult = pg_namespace_aclcheck(namespaceId,
+															  GetUserId(),
+															  ACL_CREATE);
+							if (aclresult != ACLCHECK_OK)
+								aclcheck_error(aclresult,
+										get_namespace_name(namespaceId));
+
+							renamerel(relid, stmt->newname);
+							break;
+						}
+					case RENAME_COLUMN:
+						renameatt(relid,
+								  stmt->oldname,		/* old att name */
+								  stmt->newname,		/* new att name */
+							  interpretInhOption(stmt->relation->inhOpt),		/* recursive? */
+								  false);		/* recursing already? */
+						break;
+					case RENAME_TRIGGER:
+						renametrig(relid,
+								   stmt->oldname,		/* old att name */
+								   stmt->newname);		/* new att name */
+						break;
+					case RENAME_RULE:
+						elog(ERROR, "ProcessUtility: Invalid target for RENAME: %d",
+							 stmt->renameType);
+						break;
+					default:
+						elog(ERROR, "ProcessUtility: Invalid target for RENAME: %d",
+							 stmt->renameType);
 				}
 			}
 			break;
@@ -406,6 +457,9 @@ ProcessUtility(Node *parsetree,
 		case T_AlterTableStmt:
 			{
 				AlterTableStmt *stmt = (AlterTableStmt *) parsetree;
+				Oid			relid;
+
+				relid = RangeVarGetRelid(stmt->relation, false);
 
 				/*
 				 * Some or all of these functions are recursive to cover
@@ -414,46 +468,92 @@ ProcessUtility(Node *parsetree,
 				switch (stmt->subtype)
 				{
 					case 'A':	/* ADD COLUMN */
-						AlterTableAddColumn(stmt->relname,
-										interpretInhOption(stmt->inhOpt),
+
+						/*
+						 * Recursively add column to table and, if
+						 * requested, to descendants
+						 */
+						AlterTableAddColumn(relid,
+							  interpretInhOption(stmt->relation->inhOpt),
 											(ColumnDef *) stmt->def);
 						break;
 					case 'T':	/* ALTER COLUMN DEFAULT */
-						AlterTableAlterColumnDefault(stmt->relname,
-										interpretInhOption(stmt->inhOpt),
+
+						/*
+						 * Recursively alter column default for table and,
+						 * if requested, for descendants
+						 */
+						AlterTableAlterColumnDefault(relid,
+							  interpretInhOption(stmt->relation->inhOpt),
 													 stmt->name,
 													 stmt->def);
 						break;
+					case 'N':	/* ALTER COLUMN DROP NOT NULL */
+						AlterTableAlterColumnDropNotNull(relid,
+							  interpretInhOption(stmt->relation->inhOpt),
+														 stmt->name);
+						break;
+					case 'O':	/* ALTER COLUMN SET NOT NULL */
+						AlterTableAlterColumnSetNotNull(relid,
+							  interpretInhOption(stmt->relation->inhOpt),
+														stmt->name);
+						break;
 					case 'S':	/* ALTER COLUMN STATISTICS */
-						AlterTableAlterColumnStatistics(stmt->relname,
-										interpretInhOption(stmt->inhOpt),
-														stmt->name,
-														stmt->def);
+					case 'M':	/* ALTER COLUMN STORAGE */
+
+						/*
+						 * Recursively alter column statistics for table
+						 * and, if requested, for descendants
+						 */
+						AlterTableAlterColumnFlags(relid,
+							  interpretInhOption(stmt->relation->inhOpt),
+												   stmt->name,
+												   stmt->def,
+												   &(stmt->subtype));
 						break;
 					case 'D':	/* DROP COLUMN */
-						AlterTableDropColumn(stmt->relname,
-										interpretInhOption(stmt->inhOpt),
+
+						/*
+						 * Recursively drop column from table and, if
+						 * requested, from descendants
+						 */
+						AlterTableDropColumn(relid,
+							  interpretInhOption(stmt->relation->inhOpt),
+											 false,
 											 stmt->name,
 											 stmt->behavior);
 						break;
 					case 'C':	/* ADD CONSTRAINT */
-						AlterTableAddConstraint(stmt->relname,
-										interpretInhOption(stmt->inhOpt),
+
+						/*
+						 * Recursively add constraint to table and, if
+						 * requested, to descendants
+						 */
+						AlterTableAddConstraint(relid,
+							  interpretInhOption(stmt->relation->inhOpt),
 												(List *) stmt->def);
 						break;
 					case 'X':	/* DROP CONSTRAINT */
-						AlterTableDropConstraint(stmt->relname,
-										interpretInhOption(stmt->inhOpt),
+
+						/*
+						 * Recursively drop constraint from table and, if
+						 * requested, from descendants
+						 */
+						AlterTableDropConstraint(relid,
+							  interpretInhOption(stmt->relation->inhOpt),
 												 stmt->name,
 												 stmt->behavior);
 						break;
 					case 'E':	/* CREATE TOAST TABLE */
-						AlterTableCreateToastTable(stmt->relname,
-												   false);
+						AlterTableCreateToastTable(relid, false);
 						break;
 					case 'U':	/* ALTER OWNER */
-						AlterTableOwner(stmt->relname,
-										stmt->name);
+						/* check that we are the superuser */
+						if (!superuser())
+							elog(ERROR, "ALTER TABLE: permission denied");
+						/* get_usesysid raises an error if no such user */
+						AlterTableOwner(relid,
+										get_usesysid(stmt->name));
 						break;
 					default:	/* oops */
 						elog(ERROR, "T_AlterTableStmt: unknown subtype");
@@ -464,11 +564,7 @@ ProcessUtility(Node *parsetree,
 
 
 		case T_GrantStmt:
-			{
-				GrantStmt  *stmt = (GrantStmt *) parsetree;
-
-				ExecuteGrantStmt(stmt);
-			}
+			ExecuteGrantStmt((GrantStmt *) parsetree);
 			break;
 
 			/*
@@ -483,66 +579,58 @@ ProcessUtility(Node *parsetree,
 				switch (stmt->defType)
 				{
 					case OPERATOR:
-						DefineOperator(stmt->defname,	/* operator name */
-									   stmt->definition);		/* rest */
+						DefineOperator(stmt->defnames, stmt->definition);
 						break;
 					case TYPE_P:
-						DefineType(stmt->defname, stmt->definition);
+						DefineType(stmt->defnames, stmt->definition);
 						break;
 					case AGGREGATE:
-						DefineAggregate(stmt->defname,	/* aggregate name */
-										stmt->definition);		/* rest */
+						DefineAggregate(stmt->defnames, stmt->definition);
 						break;
 				}
 			}
 			break;
 
+		case T_CompositeTypeStmt:		/* CREATE TYPE (composite) */
+			{
+				CompositeTypeStmt *stmt = (CompositeTypeStmt *) parsetree;
+
+				DefineCompositeType(stmt->typevar, stmt->coldeflist);
+			}
+			break;
+
 		case T_ViewStmt:		/* CREATE VIEW */
 			{
 				ViewStmt   *stmt = (ViewStmt *) parsetree;
 
-				DefineView(stmt->viewname, stmt->query);		/* retrieve parsetree */
+				DefineView(stmt->view, stmt->query, stmt->replace);
 			}
 			break;
 
-		case T_ProcedureStmt:	/* CREATE FUNCTION */
-			CreateFunction((ProcedureStmt *) parsetree);
+		case T_CreateFunctionStmt:		/* CREATE FUNCTION */
+			CreateFunction((CreateFunctionStmt *) parsetree);
 			break;
 
 		case T_IndexStmt:		/* CREATE INDEX */
 			{
 				IndexStmt  *stmt = (IndexStmt *) parsetree;
 
-				relname = stmt->relname;
-				if (!allowSystemTableMods && IsSystemRelationName(relname))
-					elog(ERROR, "CREATE INDEX: relation \"%s\" is a system catalog",
-						 relname);
-				if (!pg_ownercheck(GetUserId(), relname, RELNAME))
-					elog(ERROR, "permission denied");
+				CheckOwnership(stmt->relation, true);
 
-				DefineIndex(stmt->relname,		/* relation name */
+				DefineIndex(stmt->relation,		/* relation */
 							stmt->idxname,		/* index name */
 							stmt->accessMethod, /* am name */
 							stmt->indexParams,	/* parameters */
 							stmt->unique,
 							stmt->primary,
+							stmt->isconstraint,
 							(Expr *) stmt->whereClause,
 							stmt->rangetable);
 			}
 			break;
 
 		case T_RuleStmt:		/* CREATE RULE */
-			{
-				RuleStmt   *stmt = (RuleStmt *) parsetree;
-				int			aclcheck_result;
-
-				relname = stmt->object->relname;
-				aclcheck_result = pg_aclcheck(relname, GetUserId(), ACL_RULE);
-				if (aclcheck_result != ACLCHECK_OK)
-					elog(ERROR, "%s: %s", relname, aclcheck_error_strings[aclcheck_result]);
-
-				DefineQueryRewrite(stmt);
-			}
+			DefineQueryRewrite((RuleStmt *) parsetree);
 			break;
 
 		case T_CreateSeqStmt:
@@ -550,53 +638,23 @@ ProcessUtility(Node *parsetree,
 			break;
 
 		case T_RemoveAggrStmt:
-			{
-				RemoveAggrStmt *stmt = (RemoveAggrStmt *) parsetree;
-				char	   *typename = (char *) NULL;
-
-				if (stmt->aggtype != NULL)
-					typename = TypeNameToInternalName((TypeName *) stmt->aggtype);
-
-				RemoveAggregate(stmt->aggname, typename);
-			}
+			RemoveAggregate((RemoveAggrStmt *) parsetree);
 			break;
 
 		case T_RemoveFuncStmt:
-			{
-				RemoveFuncStmt *stmt = (RemoveFuncStmt *) parsetree;
-
-				RemoveFunction(stmt->funcname, stmt->args);
-			}
+			RemoveFunction((RemoveFuncStmt *) parsetree);
 			break;
 
 		case T_RemoveOperStmt:
-			{
-				RemoveOperStmt *stmt = (RemoveOperStmt *) parsetree;
-				TypeName   *typenode1 = (TypeName *) lfirst(stmt->args);
-				TypeName   *typenode2 = (TypeName *) lsecond(stmt->args);
-				char	   *typename1 = (char *) NULL;
-				char	   *typename2 = (char *) NULL;
-
-				if (typenode1 != NULL)
-					typename1 = TypeNameToInternalName(typenode1);
-				if (typenode2 != NULL)
-					typename2 = TypeNameToInternalName(typenode2);
-
-				RemoveOperator(stmt->opname, typename1, typename2);
-			}
-			break;
-
-		case T_VersionStmt:
-			elog(ERROR, "CREATE VERSION is not currently implemented");
+			RemoveOperator((RemoveOperStmt *) parsetree);
 			break;
 
 		case T_CreatedbStmt:
-			{
-				CreatedbStmt *stmt = (CreatedbStmt *) parsetree;
+			createdb((CreatedbStmt *) parsetree);
+			break;
 
-				createdb(stmt->dbname, stmt->dbpath,
-						 stmt->dbtemplate, stmt->encoding);
-			}
+		case T_AlterDatabaseSetStmt:
+			AlterDatabaseSet((AlterDatabaseSetStmt *) parsetree);
 			break;
 
 		case T_DropdbStmt:
@@ -612,7 +670,7 @@ ProcessUtility(Node *parsetree,
 			{
 				NotifyStmt *stmt = (NotifyStmt *) parsetree;
 
-				Async_Notify(stmt->relname);
+				Async_Notify(stmt->relation->relname);
 			}
 			break;
 
@@ -620,7 +678,7 @@ ProcessUtility(Node *parsetree,
 			{
 				ListenStmt *stmt = (ListenStmt *) parsetree;
 
-				Async_Listen(stmt->relname, MyProcPid);
+				Async_Listen(stmt->relation->relname, MyProcPid);
 			}
 			break;
 
@@ -628,14 +686,10 @@ ProcessUtility(Node *parsetree,
 			{
 				UnlistenStmt *stmt = (UnlistenStmt *) parsetree;
 
-				Async_Unlisten(stmt->relname, MyProcPid);
+				Async_Unlisten(stmt->relation->relname, MyProcPid);
 			}
 			break;
 
-			/*
-			 * ******************************** dynamic loader ********************************
-			 *
-			 */
 		case T_LoadStmt:
 			{
 				LoadStmt   *stmt = (LoadStmt *) parsetree;
@@ -651,14 +705,9 @@ ProcessUtility(Node *parsetree,
 			{
 				ClusterStmt *stmt = (ClusterStmt *) parsetree;
 
-				relname = stmt->relname;
-				if (IsSystemRelationName(relname))
-					elog(ERROR, "CLUSTER: relation \"%s\" is a system catalog",
-						 relname);
-				if (!pg_ownercheck(GetUserId(), relname, RELNAME))
-					elog(ERROR, "permission denied");
+				CheckOwnership(stmt->relation, true);
 
-				cluster(relname, stmt->indexname);
+				cluster(stmt->relation, stmt->indexname);
 			}
 			break;
 
@@ -667,35 +716,14 @@ ProcessUtility(Node *parsetree,
 			break;
 
 		case T_ExplainStmt:
-			{
-				ExplainStmt *stmt = (ExplainStmt *) parsetree;
-
-				ExplainQuery(stmt->query, stmt->verbose, stmt->analyze, dest);
-			}
+			ExplainQuery((ExplainStmt *) parsetree, dest);
 			break;
 
-#ifdef NOT_USED
-
-			/*
-			 * ******************************** Tioga-related statements *******************************
-			 */
-		case T_RecipeStmt:
-			{
-				RecipeStmt *stmt = (RecipeStmt *) parsetree;
-
-				beginRecipe(stmt);
-			}
-			break;
-#endif
-
-			/*
-			 * ******************************** set variable statements *******************************
-			 */
 		case T_VariableSetStmt:
 			{
 				VariableSetStmt *n = (VariableSetStmt *) parsetree;
 
-				SetPGVariable(n->name, n->args);
+				SetPGVariable(n->name, n->args, n->is_local);
 			}
 			break;
 
@@ -715,20 +743,33 @@ ProcessUtility(Node *parsetree,
 			}
 			break;
 
-			/*
-			 * ******************************** TRIGGER statements *******************************
-			 */
 		case T_CreateTrigStmt:
-			CreateTrigger((CreateTrigStmt *) parsetree);
+			CreateTrigger((CreateTrigStmt *) parsetree, false);
 			break;
 
-		case T_DropTrigStmt:
-			DropTrigger((DropTrigStmt *) parsetree);
+		case T_DropPropertyStmt:
+			{
+				DropPropertyStmt *stmt = (DropPropertyStmt *) parsetree;
+				Oid			relId;
+
+				relId = RangeVarGetRelid(stmt->relation, false);
+
+				switch (stmt->removeType)
+				{
+					case DROP_RULE:
+						/* RemoveRewriteRule checks permissions */
+						RemoveRewriteRule(relId, stmt->property,
+										  stmt->behavior);
+						break;
+					case DROP_TRIGGER:
+						/* DropTrigger checks permissions */
+						DropTrigger(relId, stmt->property,
+									stmt->behavior);
+						break;
+				}
+			}
 			break;
 
-			/*
-			 * ************* PROCEDURAL LANGUAGE statements *****************
-			 */
 		case T_CreatePLangStmt:
 			CreateProceduralLanguage((CreatePLangStmt *) parsetree);
 			break;
@@ -737,9 +778,15 @@ ProcessUtility(Node *parsetree,
 			DropProceduralLanguage((DropPLangStmt *) parsetree);
 			break;
 
+			/*
+			 * ******************************** DOMAIN statements ****
+			 */
+		case T_CreateDomainStmt:
+			DefineDomain((CreateDomainStmt *) parsetree);
+			break;
+
 			/*
 			 * ******************************** USER statements ****
-			 *
 			 */
 		case T_CreateUserStmt:
 			CreateUser((CreateUserStmt *) parsetree);
@@ -749,6 +796,10 @@ ProcessUtility(Node *parsetree,
 			AlterUser((AlterUserStmt *) parsetree);
 			break;
 
+		case T_AlterUserSetStmt:
+			AlterUserSet((AlterUserSetStmt *) parsetree);
+			break;
+
 		case T_DropUserStmt:
 			DropUser((DropUserStmt *) parsetree);
 			break;
@@ -774,11 +825,9 @@ ProcessUtility(Node *parsetree,
 			break;
 
 		case T_CheckPointStmt:
-			{
-				if (!superuser())
-					elog(ERROR, "permission denied");
-				CreateCheckPoint(false, false);
-			}
+			if (!superuser())
+				elog(ERROR, "permission denied");
+			CreateCheckPoint(false, false);
 			break;
 
 		case T_ReindexStmt:
@@ -788,32 +837,16 @@ ProcessUtility(Node *parsetree,
 				switch (stmt->reindexType)
 				{
 					case INDEX:
-						relname = (char *) stmt->name;
-						if (IsSystemRelationName(relname))
-						{
-							if (!allowSystemTableMods)
-								elog(ERROR, "\"%s\" is a system index. call REINDEX under standalone postgres with -O -P options",
-									 relname);
-							if (!IsIgnoringSystemIndexes())
-								elog(ERROR, "\"%s\" is a system index. call REINDEX under standalone postgres with -P -O options",
-									 relname);
-						}
-						if (!pg_ownercheck(GetUserId(), relname, RELNAME))
-							elog(ERROR, "%s: %s", relname, aclcheck_error_strings[ACLCHECK_NOT_OWNER]);
-						ReindexIndex(relname, stmt->force);
+						relname = (char *) stmt->relation->relname;
+						CheckOwnership(stmt->relation, false);
+						ReindexIndex(stmt->relation, stmt->force);
 						break;
 					case TABLE:
-						relname = (char *) stmt->name;
-						if (!pg_ownercheck(GetUserId(), relname, RELNAME))
-							elog(ERROR, "%s: %s", relname, aclcheck_error_strings[ACLCHECK_NOT_OWNER]);
-						ReindexTable(relname, stmt->force);
+						CheckOwnership(stmt->relation, false);
+						ReindexTable(stmt->relation, stmt->force);
 						break;
 					case DATABASE:
 						relname = (char *) stmt->name;
-						if (!allowSystemTableMods)
-							elog(ERROR, "must be called under standalone postgres with -O -P options");
-						if (!IsIgnoringSystemIndexes())
-							elog(ERROR, "must be called under standalone postgres with -P -O options");
 						ReindexDatabase(relname, stmt->force, false);
 						break;
 				}
@@ -821,10 +854,26 @@ ProcessUtility(Node *parsetree,
 			}
 			break;
 
-			/*
-			 * ******************************** default ********************************
-			 *
-			 */
+		case T_CreateConversionStmt:
+			CreateConversionCommand((CreateConversionStmt *) parsetree);
+			break;
+
+		case T_CreateCastStmt:
+			CreateCast((CreateCastStmt *) parsetree);
+			break;
+
+		case T_DropCastStmt:
+			DropCast((DropCastStmt *) parsetree);
+			break;
+
+		case T_CreateOpClassStmt:
+			DefineOpClass((CreateOpClassStmt *) parsetree);
+			break;
+
+		case T_RemoveOpClassStmt:
+			RemoveOpClass((RemoveOpClassStmt *) parsetree);
+			break;
+
 		default:
 			elog(ERROR, "ProcessUtility: command #%d unsupported",
 				 nodeTag(parsetree));
